
#include "StdAfx.h"
#include "Resource.h"
#include "Util.h"
#include "..\Dll\MouseHookDLL.h"
#include "TrayIcon.h"
#include "FoldableWindow.h"


//-----------------------------------------------------------------------
// fobOo
//-----------------------------------------------------------------------

//console_ostream dbgout;

//-----------------------------------------------------------------------
// gbvxEChE̗
//-----------------------------------------------------------------------

bool isVisibleWindow( HWND wnd )
{
	if( ::IsWindowVisible(wnd)
	 && !::IsIconic(wnd)
	 && !(WS_CHILD & ::GetWindowLong(wnd,GWL_STYLE)) )
	{
		RECT rc;
		::GetWindowRect( wnd, &rc );
		if( rc.left!=rc.right && rc.top!=rc.bottom )
			return true;
	}
	return false;
}

bool isInIgnoreList( HWND wnd )
{
	const char* ignore_list[] = {
		 "Progman",
		 "Shell_TrayWnd",
		 "tooltips_class32",
		 "SysShadow",
		 "#32768",
		 enrollUramadoClass(),
	};

	char className[20];
	::GetClassName( wnd, className, sizeof(className)-1 );

	for(int i=0; i!=sizeof(ignore_list)/sizeof(*ignore_list); ++i)
		if( 0 == strcmp(className, ignore_list[i]) )
			return true;
	return false;
}

BOOL CALLBACK EnumWindowsProc( HWND wnd, LPARAM lp )
{
	vector<HWND>& vec = *reinterpret_cast<vector<HWND>*>(lp);
	if( isVisibleWindow(wnd) && !isInIgnoreList(wnd) )
		vec.push_back( wnd );
	return TRUE;
}

void enumTopLevelWindows( vector<HWND>* vec )
{
	vec->clear();
	::EnumWindows( &EnumWindowsProc, reinterpret_cast<LPARAM>(vec) );
}

//-----------------------------------------------------------------------
// }EXCxgǐ
//-----------------------------------------------------------------------

class MouseTracer : public MouseHookHandler
{
private: // Xbh

	enum EventID { FinishThread, DragBegin, DragEnd, MouseMove, EV_NUM };
	HANDLE thread_handle;
	HANDLE mouseEvents[EV_NUM];
	CRITICAL_SECTION cs;

	void fire( EventID v ) { ::SetEvent(mouseEvents[v]); }
	static void __cdecl ThreadMain( void* p ) { static_cast<MouseTracer*>(p)->mainLoop(); }

public: // EI

	MouseTracer()
		:  left_dragging( false )
		, right_dragging( false )
	{
		for(int i=0; i!=EV_NUM; ++i)
			mouseEvents[i] = ::CreateEvent( NULL, FALSE, FALSE, NULL );
		::InitializeCriticalSection( &cs );
		thread_handle = reinterpret_cast<HANDLE>( _beginthread(&ThreadMain,0,this) );
	}
	~MouseTracer()
	{
		fire( FinishThread );
		::WaitForSingleObject( thread_handle, 1000 );
		::DeleteCriticalSection( &cs );
		for(int i=0; i!=EV_NUM; ++i)
			::CloseHandle( mouseEvents[i] );
	}

private: // CxgM

	bool  left_dragging, right_dragging;
	inline bool is_dragging() { return left_dragging | right_dragging; }

	virtual void on_event( UINT cmd, MSLLHOOKSTRUCT* dat )
	{
		bool was_dragging = is_dragging();
		::EnterCriticalSection( &cs );
			mousePt = dcomp(dat->pt.x, dat->pt.y);
		::LeaveCriticalSection( &cs );

		switch( cmd )
		{
		case WM_LBUTTONUP:
			left_dragging = false;
			if( was_dragging && !is_dragging() )
				fire( DragEnd );
			break;

		case WM_RBUTTONUP:
			right_dragging = false;
			if( was_dragging && !is_dragging() )
				fire( DragEnd );
			break;

		case WM_LBUTTONDOWN:
			left_dragging = true;
			if( !was_dragging )
				fire( DragBegin );
			break;

		case WM_RBUTTONDOWN:
			right_dragging = true;
			if( !was_dragging )
				fire( DragBegin );
			break;

		case WM_MOUSEMOVE:
			if( was_dragging )
				fire( MouseMove );
			break;
		}
	}

private: // C

	dcomp mousePt;
	vector<FoldableWindow> fwin;

	void mainLoop()
	{
		dcomp prev_pt;
		while( "infinite loop" )
		{
			DWORD v = ::WaitForMultipleObjects( EV_NUM, mouseEvents, FALSE, INFINITE );
			if( !(WAIT_OBJECT_0 <= v && v < WAIT_OBJECT_0+EV_NUM) )
				break;
			v -= WAIT_OBJECT_0;

			// XbhI
			if( v == FinishThread )
				return;

			// hbOJn
			else if( v == DragBegin )
			{
				if( fwin.empty() )
				{
					dcomp pt;
					::EnterCriticalSection( &cs );
						pt = mousePt;
					::LeaveCriticalSection( &cs );
					drag_begin( pt );
					prev_pt = pt;
				}
			}

			// hbOI
			else if( v == DragEnd )
			{
				fwin.clear();
			}

			// }EXړ
			else if( v == MouseMove )
			{
				if( !fwin.empty() )
				{
					dcomp pt;
					::EnterCriticalSection( &cs );
						pt = mousePt;
					::LeaveCriticalSection( &cs );
					if( prev_pt == pt )
						continue;

					mouse_move( prev_pt, pt );
					prev_pt = pt;
				}
			}
		}
	}

	void drag_begin( const dcomp& pt )
	{
		// 1: ŏʃEChE
		vector<HWND> hwnds;
		enumTopLevelWindows( &hwnds );
		//hwnds.resize(1);

		// 2: GbW߂drag͐܂Ȃ[hJnȂi^Cgo[hbOetcj
		POINT ptx = {long(pt.real()), long(pt.imag())};
		for(int i=0; i!=hwnds.size(); ++i)
		{
			RECT rc; ::GetWindowRect( hwnds[i], &rc );
			if( ::PtInRect(&rc, ptx) )
			{
				::InflateRect( &rc, -::GetSystemMetrics(SM_CXVSCROLL), -::GetSystemMetrics(SM_CYCAPTION) );
				if( !::PtInRect(&rc, ptx) )
					return;
				break;
			}
		}

		// 3: eXɑ΂ėnEChE쐬
		vector<HWND> huras( hwnds.size() );
		transform( hwnds.begin(), hwnds.end(), huras.begin(), &GenerateUraWnd );

		// 4: ZI[_[
		HWND prev = HWND_TOP;
		for(int i=0; i!=huras.size(); ++i)
			::SetWindowPos( huras[i], prev, 0, 0, 0, 0, SWP_NOSIZE|SWP_NOMOVE );

		// 5: FoldableWindow쐬
		fwin.resize( hwnds.size() );
		for(int i=0; i!=fwin.size(); ++i)
			fwin[i].Init( hwnds[i], huras[i] );
	}

	void mouse_move( const dcomp& prev, const dcomp& pt )
	{
		// 1: SĂ̑Ƀ}EXCxgM
		int j = -1;
		bool no_mekuri = false;
		for(int i=0; i!=fwin.size(); ++i)
		{
			if( no_mekuri && fwin[i].is_taira() ) continue;
			if( fwin[i].mouse( prev, pt ) )
				j = i;
			else if( !no_mekuri && fwin[i].covers(pt) ) // ȏ̑͌ĂȂ̂ŃCxgMȂ
				no_mekuri = true;
		}

		// ToDo: ReBendKv͍̂XVƂFœK
		// ToDo: ĕ`͍Ōɂ܂Ƃ߂悤

		// 2: Đ܂Ȃ
		for(int i=j; i>=0; --i)
		{
			dcomp c1, c2;
			if( fwin[i].get_orime(c1,c2) )
				for(int k=i-1; k>=0; --k)
					fwin[k].rebend(c1,c2);
		}
	}
};

//-----------------------------------------------------------------------
// gCj[
//-----------------------------------------------------------------------

class OriMadoMenu : public TrayIconEventHandler
{
public:
	OriMadoMenu() : menu( load_menu(IDM_MAINMENU) ) {}
	~OriMadoMenu() { ::DestroyMenu( menu ); }

private:
	const HMENU menu;
	virtual HMENU on_menu() { return ::GetSubMenu( menu, 0 ); }
	virtual void  on_command( HMENU, UINT id )
	{
		switch( id )
		{
		case ID_CMD_EXIT:
			::PostQuitMessage( 0 );
			break;
		case ID_CMD_CONFIG:
			msgbox("Under Construction");
			break;
		}
	}
};

//-----------------------------------------------------------------------
// C
//-----------------------------------------------------------------------

int APIENTRY WinMain( HINSTANCE, HINSTANCE, LPSTR, int )
{
	// ^XNgCɏ풓
	OriMadoMenu OriMadoMenu;
	TrayIcon tsi( &OriMadoMenu, load_string(IDS_APP_TITLE).c_str() );
	tsi.setIcon( load_icon(IDI_MAINICON) );
	tsi.setTip ( load_string(IDS_APP_TITLE) );

	// }EXĎ
	MouseTracer mt;
	MouseHookBegin( &mt );
	for( MSG msg; ::GetMessage(&msg,NULL,0,0); )
	{
		::TranslateMessage( &msg );
		::DispatchMessage( &msg );
	}
	MouseHookEnd();
}
